open Object
open Syntax
open List
open Util

let rec compile_exp exp =
  match exp with
    | Varref ident -> [O_VRef ident]
    | Call(f,args,None) ->
        (concat (rev_map compile_exp args)) @
          (compile_exp f) @ [O_Call((length args),false)]
    | Call(f,args,Some rest_arg) ->
        (compile_exp rest_arg)
        @ (concat (rev_map compile_exp args)) 
        @ (compile_exp f) @ [O_ExCall(((length args)+1),false)]
    | IntLit i -> [O_IntLit i]
    | String str -> [O_StrLit str]
    | SGet(exp, ident) ->
        (compile_exp exp) @ [O_SGet ident]
    | SSet(exp1, ident, exp2) ->
        (compile_exp exp1) @ (compile_exp exp2) @ [O_SSet ident]
    | Assign(ident, exp) ->
        (compile_exp exp) @ [O_Assign ident]
    | MakeFun(params, rparam, body) ->
        [O_MkFun(params, rparam, compile_exps body)]
    | IfStmt(exp,then_bd,else_bd) ->
        let cond_ops = compile_exp exp in
        let then_ops = compile_exps then_bd in
        let else_ops = compile_exps else_bd in
          cond_ops @
            [O_Branch(0, (length then_ops) + 1)] @
            then_ops @
            [O_Goto (length else_ops)] @
            else_ops
    | And(exp1, exp2) ->
        let cond_ops = compile_exp exp1 in
        let then_ops = compile_exp exp2 in
          cond_ops @
            [O_Dup; O_Branch(0, (length then_ops)+1); O_Pop] @
            then_ops
    | Or(exp1, exp2) ->
        let cond_ops = compile_exp exp1 in
        let else_ops = compile_exp exp2 in
          cond_ops @
            [O_Dup; O_Branch((length else_ops)+1, 0); O_Pop] @
            else_ops
    | Vardef(ident, exp) -> (compile_exp exp) @ [O_VDef ident]
    | Not(exp) ->
        (compile_exp exp) @ [O_Push fun_obj_not;O_Call(1,false)]
    | MakeList exps ->
        (concat (rev_map compile_exp exps)) @
          [O_Push new_list; O_Call((length exps),false)]
    | Class(ident, super_exp, exps) ->
        let super_ops =
          match super_exp with
            | None -> [O_VRef "Object"]
            | Some e -> compile_exp e
        in
        let set_super =
          ([O_VRef "self"] @ super_ops @ [O_SSet "_super_"; O_Pop])
        in
        let set_methods =
          [O_VRef "self"; O_Push (make_native_fun_0 new_obj);
           O_Call (0,false); O_SSet "_methods_"; O_Pop]
        in
        let set_class =
          [O_VRef "self"; O_Push class_Class; O_SSet "_class_"; O_Pop]
        in
          [O_Push (make_native_fun_0 plain_obj);
           O_Call (0,false);
           O_VDef ident;
           O_MkFun(["self"], None,
                   set_super @ set_methods @ set_class @ (compile_exps exps));
           O_Call (1,false)]
    | Return ((Call _) as exp) -> begin
        match part_tail (compile_exp exp) with
          | (ops, O_Call(n,_)) -> ops @ [O_Call(n,true)]
          | (ops, O_ExCall(n,_)) -> ops @ [O_ExCall(n,true)]
      end
    | Return exp ->
        (compile_exp exp) @ [O_VRef "*return*"; O_Call(1,false)]
        
and compile_exps = function
  | [] -> [O_Push NilObj]
  | exp::[] -> compile_exp exp
  | exp::rest -> (compile_exp exp) @ [O_Pop] @ (compile_exps rest)
  
