open Object
open Syntax
open List
open Printf
open Util
open Debug

let rec search_var ident context =
  match context with
    | ToplevelContext -> None
    | Context (outer, table) ->
        if Hashtbl.mem table ident then
          Some table
        else
          search_var ident outer
            
let rec stack_size env =
  match (search_var "*return*" env) with
    | None -> 0
    | Some table ->
        match Hashtbl.find table "*return*" with
          | ContObj(old_env,_,_) -> 1 + (stack_size old_env)
          | _ -> bug_alert "*return* must be continuation object"

let var_get ident context =
  match (search_var ident context) with
    | None -> raise_error (sprintf "no such variable %s" ident)
    | Some table -> Hashtbl.find table ident

let var_set ident context obj =
  match (search_var ident context) with
    | None -> raise_error "no such variable"
    | Some table -> Hashtbl.replace table ident obj

let var_def ident context obj =
  match context with
    | ToplevelContext -> bug_alert "refer toplevel context"
    | Context (_, table) -> Hashtbl.replace table ident obj

let get_slot_with_check obj slot =
  match (get_slot obj slot) with
    | Undef -> raise_error (Printf.sprintf "%s has no slot [%s]"
                              (show_obj obj) slot)
    | obj -> obj
                              
        
let rec eval_ops ops env vstack =
  let funcall nargs rops tail =
    let arity_error n m =
      raise_error (Printf.sprintf "wrong number of arguments (%d for %d)" n m)
    in
      match (hd vstack) with
        | FunObj (outer, params, None, body) ->
            let nparams = length params in
              if nargs = nparams  then
                let return =
                  if tail then
                    var_get "*return*" env
                  else
                    ContObj(env, (remove (nargs+1) vstack), rops)
                in
                let param_bind = combine params (take nargs (tl vstack)) in
                let new_table  = 
                  create_table (("*return*",return)::param_bind)
                in
                  eval_ops
                    body
                    (Context(outer, new_table ))
                    (remove (nargs+1) vstack)
              else
                arity_error nargs nparams
        | FunObj (outer, params, Some rparam, body) ->
            let nparams = length params in
              if nparams <= nargs then
                let norm_bind = combine params (take nparams (tl vstack)) in
                let rest_args = take (nargs-nparams) (remove (nparams+1) vstack) in
                let rest_bind = (rparam, ListObj rest_args) in
                let return =
                  if tail then
                    var_get "*return*" env
                  else
                    ContObj(env, (remove (nargs+1) vstack), rops)
                in
                let return_bind =
                  ("*return*", return)
                in
                let new_table =
                  create_table (return_bind::rest_bind::norm_bind)
                in
                  eval_ops
                    body
                    (Context(outer, new_table ))
                    (remove (nargs+1) vstack)
        | BoundFunObj (arity, rp, boundobjs, func) ->
            if arity = nargs || (rp && arity <= nargs) then
              eval_ops
                ([O_Call(((length boundobjs)+nargs),tail)] @ rops)
                env
                ([func] @ boundobjs @ (tl vstack))
            else
              arity_error nargs arity
        | NativeFunObj (arity,va_arg,func) ->
            if nargs = arity || (va_arg && nargs >= arity ) then
              eval_ops
                rops
                env
                ((func (take nargs (tl vstack)))::(remove (nargs+1) vstack))
            else
              arity_error nargs arity
        | CallCCObj ->
            if nargs = 1 then
              let cont = ContObj(env, (remove 2 vstack), rops) in
              let func = nth vstack 1 in
                eval_ops
                  (O_Call(1,tail)::rops)
                  env
                  (func::cont::(remove 2 vstack))
            else
              arity_error nargs 1
        | ContObj(old_env, old_vstack, old_ops) ->
            if nargs = 1 then
              eval_ops
                old_ops
                old_env
                ((nth vstack 1)::old_vstack)
            else
              arity_error nargs 1
        | _ -> raise_error ("must be function but " ^ (show_obj (hd vstack)))
  in
  (* let _ = (if ops != [] then print_endline (show_op (hd ops))) in  *)
  (* let _ = print_objs vstack in *)
  (* let _ = (print_int (stack_size env); print_endline "") in *)
    match ops with
      | [] -> begin
          match (var_get "*return*" env) with
            | ContObj(old_env, old_vstack, old_ops) ->
                eval_ops old_ops old_env ((List.hd vstack)::old_vstack)
            | _ -> raise_error "no continuation"
        end
      | O_Stop::r -> ()
      | O_Pop::r -> eval_ops r env (List.tl vstack)
      | O_Dup::r -> eval_ops r env ((hd vstack)::vstack)
      | O_Nop::r -> eval_ops r env vstack
      | (O_Push obj)::r -> eval_ops r env (obj::vstack)
      | (O_IntLit i)::r -> eval_ops r env ((IntObj i)::vstack)
      | (O_StrLit st)::r -> eval_ops r env ((StringObj st)::vstack)
      | (O_VRef ident)::r -> 
          eval_ops r env ((var_get ident env)::vstack)
      | (O_VDef ident)::r -> begin
          var_def ident env (hd vstack);
          eval_ops r env vstack;
        end
      | (O_Call(nargs,tail))::r -> funcall nargs r tail
      | (O_ExCall(nargs,tail))::r -> 
          let top = take nargs vstack in
          let list = nth vstack nargs in
          let bottom = remove (nargs+1) vstack in begin
            match list with
                ListObj l ->
                  eval_ops
                    ((O_Call((length l)+nargs-1,tail))::r)
                    env
                    (top @ l @ bottom)
              | _ -> raise_error "must be list, list extraction call"
            end
      | (O_SGet ident)::r ->
          eval_ops
            r
            env
            ((get_slot_with_check (hd vstack) ident)::(tl vstack) )
      | (O_SSet ident)::r -> begin
          set_slot (nth vstack 1) ident (hd vstack);
          eval_ops r env ((hd vstack)::(remove 2 vstack));
        end
      | (O_Assign ident)::r -> begin
          var_set ident env (hd vstack);
          eval_ops r env vstack;
        end
      | (O_MkFun (params, rparam, ops))::r ->
          eval_ops
            r
            env
            (FunObj(env, params, rparam, ops)::vstack)
      | (O_Goto n)::r ->
          eval_ops (remove n r) env vstack
      | (O_Branch(n, m))::r ->
          if test (hd vstack) then
            eval_ops (remove n r) env (tl vstack)
          else
            eval_ops (remove m r) env (tl vstack)
let new_context () =
  let table = create_table [
    ("print_debug", (make_native_fun_1 print_debug));
    ("Object", class_Object);
    ("Class", class_Class);
    ("Integer", class_Integer);
    ("Function", class_Function);
    ("List", class_List);
    ("TrueClass", class_True);
    ("FalseClass", class_False);
    ("Continuation", class_Continuation);
    ("callcc", CallCCObj);
    ("bind", NativeFunObj(1,true, fun args -> bind (hd args) (tl args)));
    ("not", fun_obj_not);
    ("true", TrueObj);
    ("false", FalseObj);
    ("nil", NilObj)
  ]
  in
    Context (ToplevelContext, table)
      

