open Syntax
open Debug
open Printf
open List
open Util
open Debug
  
let print_debug obj =
  begin
    print_endline ("#" ^ (show_obj obj));
    NilObj;
  end

(* convert ocaml function to twilras object *)
let make_native_fun_0 f =
  let wrapped_f = function
      [] -> f ()
    | _ -> bug_alert "bug: number of argument"
  in
    NativeFunObj(0,false, wrapped_f)
      
let make_native_fun_1 f =
  let wrapped_f = function
      arg::[] -> f arg
    | _ -> bug_alert "bug: number of argument"
  in
    NativeFunObj(1, false, wrapped_f)

let make_native_fun_2 f =
  let wrapped_f = function
      arg1::arg2::[] -> f (arg1, arg2)
    | _ -> bug_alert "bug: number of argument"
  in
    NativeFunObj(2, false, wrapped_f)

(* basic functions *)
let test = function
  | NilObj
  | FalseObj -> false
  | _ -> true

let boolToObj = function
  | true -> TrueObj
  | false -> FalseObj
    
let create_table pairs =
  let table = Hashtbl.create 20 in
  let add_entry = function
      (key, data) -> Hashtbl.replace table key data
  in
    begin
      List.iter add_entry pairs;
      table
    end

let fun_obj_not = NativeFunObj(1,false, fun a -> boolToObj(not (test (hd a))))
  
(* bind object to function *)
let bind func objs =
  match func with
    | NativeFunObj (arity, va_args, f) ->
        if ((length objs) <= arity) || va_args then
          NativeFunObj (max 0 (arity-(length objs)),
                        va_args,
                        fun args -> f (objs @ args))
        else
          raise_error "too many args, cannot bind"
    | FunObj(context, params, None, body) ->
        let n_restarg = (length params) - (length objs) in
          if n_restarg >= 0 then
            BoundFunObj(n_restarg, false, objs, func)
          else
            raise_error "too many args, cannot bind"
    | FunObj(_, params, Some rparam, _) ->
        BoundFunObj(max 0 ((length params) - (length objs)), true, objs, func)
    | BoundFunObj(arity, false, boundobjs, func) ->
        let n_restarg = arity - (length objs) in
          if n_restarg >= 0 then
            BoundFunObj(n_restarg, false, (boundobjs @ objs), func)
          else
            raise_error "too many args, cannot bind"
    | BoundFunObj(arity, true, boundobjs, func) ->
        BoundFunObj(max 0 (arity-(length objs)), true, (boundobjs @ objs), func)
    | CallCCObj ->
        if (length objs) = 0 then
          CallCCObj
        else if (length objs) = 1 then
          BoundFunObj(0, false, objs, CallCCObj)
        else
          raise_error "too many args, cannot bind"
    | _ -> raise_error "cannot bind to non function object"

(* class Object *)
let plain_obj () = NormObj(create_table [])
  
let class_Object =
  let methods = create_table [
    ("initialize", NativeFunObj(0,true, fun a -> NilObj));
    ("nilp", NativeFunObj(1, false, fun a -> FalseObj));
  ]
  in
  let slots = create_table [
    ("_methods_", NormObj methods);
    ("plain", make_native_fun_0 plain_obj);
  ]
  in
    NormObj (slots)
let new_obj () = NormObj(create_table [("_class_",class_Object)])
let _ = match class_Object with
    NormObj slots -> Hashtbl.replace slots "new" (make_native_fun_0 new_obj)
  | _ -> bug_alert "????"

(* class Class *)
let class_Class =
  let new_ops =
    [ O_Push (make_native_fun_0 new_obj);
      O_Call (0,false);
      O_VDef "newobj"; (* newobj = Object.plain() *)
      O_VRef "self";
      O_SSet "_class_";
      O_Pop; (* newobj._class_ = self *)
      O_VRef "args";
      O_VRef "newobj";
      O_SGet "initialize";
      O_ExCall (1,false);
      O_Pop; (* newobj.initialize( *args ) *)
      O_VRef "newobj";
    ]
  in
  let new_fun = FunObj(ToplevelContext, ["self"], Some "args", new_ops) in
  let methods = create_table [("new",new_fun)] in
  let slots = create_table [
    ("_super_",class_Object);
    ("_methods_", NormObj methods);
  ] in
    NormObj slots

let _ = match class_Object with
    NormObj slots -> Hashtbl.replace slots "_class_" class_Class
  | _ -> bug_alert "????"

let _ = match class_Class with
    NormObj slots -> Hashtbl.replace slots "_class_" class_Class
  | _ -> bug_alert "????"
      
(* class Integer *)
let class_Integer =
  let obj_of_fun_int_int func =
    let f = function
      | (IntObj i) -> IntObj(func i)
      | _ -> raise_error "must be (int,int)"
    in
      make_native_fun_1 f
  in
  let obj_of_fun_int2_int func =
    let f = function
      | (IntObj i, IntObj j) -> IntObj(func i j)
      | _ -> raise_error "must be (int,int)"
    in
      make_native_fun_2 f
  in
  let obj_of_fun_int2_bool func =
    let f = function
      | (IntObj i, IntObj j) -> boolToObj (func i j)
      | _ -> raise_error "must be (int,int)"
    in
      make_native_fun_2 f
  in
  let equal = function
    | (IntObj i, IntObj j) -> boolToObj (i=j)
    |  _ -> FalseObj
  in
  let methods = create_table [
    ("+", obj_of_fun_int2_int (+));
    ("-", obj_of_fun_int2_int (-));
    ("*", obj_of_fun_int2_int ( * ));
    ("/", obj_of_fun_int2_int (/));
    ("==", make_native_fun_2 equal);
    ("-@", obj_of_fun_int_int (~-));
    ("abs", obj_of_fun_int_int abs);
    ("<", obj_of_fun_int2_bool (<));
    ("<=", obj_of_fun_int2_bool (<=));
    (">=", obj_of_fun_int2_bool (>=));
    (">", obj_of_fun_int2_bool (>));
  ]
  in
  let slots = create_table [
    ("_super_", class_Object);
    ("_methods_", NormObj(methods));
  ]
  in
    NormObj slots

(* class List *)
let new_list = NativeFunObj(0, true, fun objs -> (ListObj objs))
let class_List =
  let hd = function
    | ListObj l -> (List.hd l)
    | _ -> raise_error "must be list, hd"
  in
  let tl = function
    | ListObj l -> ListObj (List.tl l)
    | _ -> raise_error "must be list, tl"
  in
  let empty_p = function
    | ListObj l -> boolToObj (l=[])
    | _ -> raise_error "must be list, tl"
  in
  let methods = create_table [
    ("hd", (make_native_fun_1 hd));
    ("tl", (make_native_fun_1 tl));
    ("empty_p", (make_native_fun_1 empty_p));
  ]
  in
  let empty () = (ListObj []) in
  let cons = function
    | (obj, ListObj l) -> ListObj (obj::l)
    | _ -> raise_error "must be list,cons"
  in
  let slots = create_table [
    ("_super_", class_Object);
    ("_methods_", NormObj(methods));
    ("empty", make_native_fun_0(empty));
    ("cons", make_native_fun_2(cons));
    ("new", new_list);
  ]
  in
    NormObj slots

(* class Function *)
let class_Function =
  let methods = create_table [
    ("bind", NativeFunObj(1, true, fun args -> bind (hd args) (tl args)));
  ]
  in
  let slots = create_table [
    ("_class_", class_Class);
    ("_super_", class_Object);
    ("_methods_", NormObj methods);
    ("new", Undef);
  ]
  in
    NormObj slots

(* class TrueClass *)
let class_True =
  let slots = create_table [
    ("_class_", class_Class);
    ("_super_", class_Object);
    ("_methods_", NormObj (create_table []));
    ("new", Undef);
  ]
  in
    NormObj slots

(* class FalseClass *)
let class_False =
  let slots = create_table [
    ("_class_", class_Class);
    ("_super_", class_Object);
    ("_methods_", NormObj (create_table []));
  ]
  in
    NormObj slots

(* class NilClass *)
let class_Nil =
  let methods = create_table [
    ("nilp", NativeFunObj(1, false, fun a -> TrueObj));
  ]
  in
  let slots = create_table [
    ("_class_", class_Class);
    ("_super_", class_Object);
    ("_methods_", NormObj methods);
    ("new", Undef);
  ]
  in
    NormObj slots

(* class Continuation *)
let class_Continuation =
  let ops_call = [
    O_VRef "ret";
    O_VRef "self";
    O_Call (1,false);
  ]
  in
  let methods = create_table [
    ("call", FunObj(ToplevelContext, ["self"; "ret"], None, ops_call));
  ]
  in
  let slots = create_table [
    ("_class_", class_Class);
    ("_super_", class_Object);
    ("_methods_", NormObj methods);
    ("new", Undef);
  ]
  in
    NormObj slots

(* slot access function *)
let rec get_method _class obj key =
  match _class with
    | NormObj slots -> 
        if Hashtbl.mem slots "_methods_" then
          let f =  get_slot (Hashtbl.find slots "_methods_") key in
            match f with
              | FunObj _
              | NativeFunObj _ -> bind f [obj]
              | Undef ->
                  if Hashtbl.mem slots "_super_" then
                    get_method (Hashtbl.find slots "_super_") obj key
                  else
                    Undef
              | _ -> Undef 
        else
          raise_error "must have _methods_ slot" 
    | _ -> Undef
and get_slot obj key =
  match obj with
    | NormObj slots ->
        if Hashtbl.mem slots key then
          Hashtbl.find slots key
        else if Hashtbl.mem slots "_class_" then
          get_method (Hashtbl.find slots "_class_") obj key
        else
          Undef
    | IntObj _ -> get_method class_Integer obj key
    | ListObj _ -> get_method class_List obj key
    | FunObj _
    | BoundFunObj _ 
    | NativeFunObj _ -> get_method class_Function obj key
    | TrueObj -> get_method class_True obj key
    | FalseObj -> get_method class_False obj key
    | NilObj -> get_method class_Nil obj key
    | ContObj _ -> get_method class_Continuation obj key
    | CallCCObj -> get_method class_Function obj key
    | Undef -> bug_alert "bug! refer Undef"

let set_slot obj key v =
  match obj with
    | NormObj slots ->
        Hashtbl.replace slots key v;
    | _ -> raise_error (sprintf "cannot set slot to %s" (show_obj obj))
